#!/usr/bin/env python

import re
import subprocess
from pathlib import Path


repo_root = Path(__file__).resolve().parent.parent
pyproject_path = repo_root / "pyproject.toml"


def ask_bump_type() -> int:
    while True:
        print("What kind of bump do you want?\n\n  (1) patch\n  (2) minor\n  (3) major\n")
        answer = input("[Default 1]: ")
        if answer == "":
            answer = 1
        else:
            try:
                answer = int(answer)
            except ValueError:
                continue
        break

    return answer


def get_new_version(current_version: str) -> str:
    print(f"Current version: {current_version}")
    bump_type = ask_bump_type()
    print()

    version_matches = re.search(r"(\d+)\.(\d+)\.(\d+)", current_version)
    if not version_matches:
        print("Current version looks funky, please fix before you continue")
        exit(1)
    cur_major, cur_minor, cur_patch = version_matches.groups()

    if bump_type == 1:
        cur_patch = int(cur_patch) + 1
    elif bump_type == 2:
        cur_minor = int(cur_minor) + 1
        cur_patch = 0
    elif bump_type == 3:
        cur_major = int(cur_major) + 1
        cur_patch = cur_minor = 0

    new_version = f"{cur_major}.{cur_minor}.{cur_patch}"
    print(f"\nnew version: {new_version}")
    if input("Does that look ok? [Yn]: ") == "n":
        exit(1)

    return new_version


def commit_version_change(version: str):
    answer = input("Do you want to commit the version bump and create a new tag? [Yn]")
    if answer == "n":
        return

    commit_msg = f"version: bump to {version}"
    tag_msg = f"Version {version}"
    output = subprocess.check_output(["git", "-P", "diff", "--name-only", "--cached"]).decode("utf-8")
    if output != "":
        print("The following files are staged, which is not expected.\n")
        print(output)
        print(
            "Make sure you have no staged files before using this script. "
            "For example by running 'git restore --staged -- .'"
        )
        exit(1)

    subprocess.run(
        ["git", "add", pyproject_path],
        check=True,
    )
    subprocess.run(["git", "commit", "-m", commit_msg])

    subprocess.run(["git", "tag", "-a", version, "-m", tag_msg])

    print("Commit built, dont forget to push :)")
    print(f'To push the new created tag use the following command: "git push origin tag {version}"')


def get_current_version(pyproject_lines: list[str]) -> str:
    current_version = None
    for line in pyproject_lines:
        if line.startswith("version"):
            current_version = line.split("=")[1].strip()

    if not current_version:
        print('Found no line starting with "version" in the pyproject.toml. Please fix before you continue')
        exit(1)

    return current_version


def write_version(version: str, pyproject_lines: list[str]):
    with open(pyproject_path, "w") as pyproject_file:
        for line in pyproject_lines:
            if line.startswith("version"):
                line = f'version = "{version}"\n'
            pyproject_file.write(line)


def main():
    with open(pyproject_path, "r") as pyproject_file:
        pyproject_lines = pyproject_file.readlines()

    current_version = get_current_version(pyproject_lines)
    new_version = get_new_version(current_version)
    write_version(new_version, pyproject_lines)

    commit_version_change(new_version)


if __name__ == "__main__":
    main()



# coding=utf-8
import re


def strip_duration_and_seconds(timer_msg: str) -> str:
    """
    For a string like:
        "Finish 'cross-compile doom' in 0:00:00.000002 (0.000002s total)"

    this replaces the wobbly time parts, so it becomes:
        "Finish 'cross-compile doom' in <duration>"
    """

    timer_msg = re.sub(r"(.* in )0:00:.*", r"\1<duration>", timer_msg)
    return timer_msg



from datetime import date
from unittest import TestCase

from baumbelt.grouping import group_by_key


class GroupByKeyTestCase(TestCase):
    def test_group_by_callable_key(self):
        iterable = [
            date(2020, 1, 1),
            date(2021, 2, 2),
            date(2022, 3, 3),
            date(2023, 4, 4),
        ]

        grouped = group_by_key(iterable, "weekday")
        self.assertEqual(
            grouped,
            {
                1: [date(2021, 2, 2), date(2023, 4, 4)],
                2: [date(2020, 1, 1)],
                3: [date(2022, 3, 3)],
            },
        )

    def test_group_by_key(self):
        iterable = [
            date(2020, 1, 1),
            date(2021, 2, 2),
            date(2022, 3, 3),
            date(2023, 4, 4),
        ]

        grouped = group_by_key(iterable, "day")
        self.assertEqual(
            grouped,
            {
                1: [date(2020, 1, 1)],
                2: [date(2021, 2, 2)],
                3: [date(2022, 3, 3)],
                4: [date(2023, 4, 4)],
            },
        )



import contextlib
import time

import pygments
import sqlparse
from django.core.management.color import supports_color
from django.db import connections
from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers import SqlLexer
from sqlparse.sql import Parenthesis, IdentifierList, Identifier

MAX_CHARS_TO_PARSE = 5_000
UNPARSED_CHARS_TO_PRINT = 400


class DjangoSQLWrapper:
    def __init__(self, indent=False, max_arguments=5, truncate_unparsable=True):
        self.do_indent = indent
        self.max_arguments = max_arguments
        self.truncate_unparsable = truncate_unparsable

    def __call__(self, execute, sql, params, many, context):
        start = time.monotonic()
        try:
            return execute(sql, params, many, context)
        except Exception as e:
            raise e
        finally:
            duration = time.monotonic() - start
            sql_log = self.log_executemany(sql, params) if many else self.log_execute(sql % params)
            print(f"({duration:.2f}s)\n{sql_log}")

    def truncate_arguments(self, sql):
        parsed = sqlparse.parse(sql)
        modified_statements = []

        for statement in parsed:
            stmt = self._trim_in(statement.tokens)
            modified_statements.append("".join([str(token) for token in stmt]))

        return " ".join(modified_statements)

    def _trim_in_arguments(self, parenthesis_token):
        arguments = parenthesis_token.value.strip("()").split(",")
        if len(arguments) <= self.max_arguments or len(arguments) < 3:
            return parenthesis_token

        last_arg = arguments[-1]
        next_n_args = arguments[: self.max_arguments - 1]

        # When indenting, this represents the amount of spaces to the left.
        indentation = len(last_arg) - len(last_arg.strip()) - 1

        args_removed = len(arguments) - len(next_n_args) - 1
        truncation = f"{indentation * ' '}/* {args_removed} truncated */"

        if self.do_indent:
            truncation = f"\n{truncation}"

        first_args_str = ",".join(next_n_args)
        new_value = f"({first_args_str},{truncation}{last_arg})"
        return Parenthesis(sqlparse.sql.TokenList(sqlparse.parse(new_value)[0].tokens))

    def _trim_in(self, tokens):
        """
        Looks for an IN expression, and truncates its arguments to max_items.

        ```
        WHERE "table"."id" IN (0,1,2, ..., 999)
        ```
        """

        for i, token in enumerate(tokens):
            if token.ttype is sqlparse.tokens.Keyword and token.value.upper() == "IN":
                # The next two tokens should be Whitespace followed by Parenthesis.
                next_token_parenthesis = tokens[i + 2]

                if isinstance(next_token_parenthesis, Parenthesis):
                    truncated_arguments = self._trim_in_arguments(next_token_parenthesis)
                    tokens[i + 2] = truncated_arguments

            if isinstance(token, (IdentifierList, Identifier)):
                self._trim_in(token.tokens)

            if token.is_group:
                self._trim_in(token.tokens)

        return tokens

    def log_executemany(self, sql, param_list):
        log = ""
        for param_tuple in param_list:
            log += f"{self.log_execute(sql % param_tuple)}\n"
        return log

    def log_execute(self, sql):
        sql_length = len(sql)
        if sql_length > MAX_CHARS_TO_PARSE:
            if self.truncate_unparsable:
                sql = f"{sql[:UNPARSED_CHARS_TO_PRINT]} <{sql_length - UNPARSED_CHARS_TO_PRINT} chars hidden>"

            return sql

        formatted = sqlparse.format(sql, reindent=self.do_indent)
        if self.max_arguments >= 0:
            truncated = self.truncate_arguments(sql=formatted)
        else:
            truncated = formatted

        if supports_color():
            pretty_sql = pygments.highlight(truncated, SqlLexer(), TerminalFormatter())
        else:
            pretty_sql = truncated

        return pretty_sql


@contextlib.contextmanager
def django_sql_debug(
        indent: bool = False,
        max_arguments: int = 5,
        truncate_unparsable: bool = True,
        db_name: str = "default"
):
    dj_sql_wrap = DjangoSQLWrapper(indent=indent, max_arguments=max_arguments, truncate_unparsable=truncate_unparsable)
    with connections[db_name].execute_wrapper(dj_sql_wrap):
        yield



import inspect
import time
from datetime import timedelta, datetime
from typing import Literal

try:
    from django.db import connections
    from django.db.backends.dummy.base import DatabaseWrapper

    DJANGO_AVAILABLE = True
except ImportError:
    DJANGO_AVAILABLE = False


UNSET = object()


class MeasureTime:
    duration: timedelta
    start: datetime

    def __enter__(self):
        self.start = datetime.now()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.duration = datetime.now() - self.start

    def __str__(self):
        if self.duration is None:
            self.duration = datetime.now() - self.start
        return f"{self.duration} ({self.duration.total_seconds()}s)"


class Timer:
    call_stacks = {}
    last_tap = None
    last_tap_queries = 0

    creset = "\33[0m"
    cgreen = "\33[32m"
    cgrey = "\33[90m"

    def __init__(
        self,
        name: str = None,
        resolution: Literal["s"] | Literal["ms"] = "s",
        disable: bool = False,
        db_name: str = "default",
        disable_queries: bool = UNSET,
    ):
        if name is None:
            _, _, _, function_name, _, _ = inspect.stack()[1]
            name = function_name

        self.name = name
        self.resolution = resolution
        self.disable = disable
        self.db_name = db_name

        # If not specified, use them if django is available
        if disable_queries is UNSET:
            disable_queries = not DJANGO_AVAILABLE

        # Fail if mentioned explicitly, but Django is not present.
        elif disable_queries is False and not DJANGO_AVAILABLE:
            raise ValueError(
                "Queries can't be tracked if Django is not installed. Hint: remove your `disable_queries` argument."
            )

        self.enable_queries = not disable_queries

    def _get_padding(self):
        return " " * (Timer.call_stacks[self]) * 2

    def _convert_to_resolution(self, delta) -> tuple[int, str]:
        if self.resolution == "ms":
            return delta * 1000, self.resolution

        return delta, self.resolution

    def __enter__(self):
        Timer.call_stacks[self] = len(Timer.call_stacks.keys())

        if self.disable:
            return self

        if self.enable_queries:
            con: DatabaseWrapper = connections[self.db_name]
            self.queries_start = len(con.queries)
            self.last_tap_queries = self.queries_start

        msg = f"{self._get_padding()}v '{self.name}' started..."
        print(f"{self.cgreen}{msg}{self.creset}")
        self.start = time.time()
        self.last_tap = self.start

        return self

    def tap(self, name: str):
        now = time.time()
        last_tap = self.last_tap
        self.last_tap = now
        if self.disable:
            return

        padded_tap_name = f"'{name}'".ljust(30)
        msg = f"{self._get_padding()} > {padded_tap_name}"
        msgs = []

        delta = now - last_tap
        delta, unit = self._convert_to_resolution(delta)
        duration, _ = self._convert_to_resolution(now - self.start)

        msgs.append(f"took {delta:.4f}{unit} (at {duration:.4f}{unit})")

        if self.enable_queries:
            con: DatabaseWrapper = connections[self.db_name]
            queries_done = len(con.queries)
            amount_queries = queries_done - self.last_tap_queries
            self.last_tap_queries = queries_done
            msgs.append(f"had {amount_queries} queries")

        msg = msg + ", ".join(msgs)
        print(f"{self.cgrey}{msg}{self.creset}")

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.disable:
            del Timer.call_stacks[self]
            return

        msg = f"{self._get_padding()}ʌ '{self.name}'"
        end = time.time()
        duration = end - self.start
        self.total = duration
        duration, unit = self._convert_to_resolution(duration)
        msg += f" took {duration:.4f}{unit},"

        if self.enable_queries:
            con: DatabaseWrapper = connections[self.db_name]
            queries_done = len(con.queries) - self.queries_start
            msg += f" had {queries_done} queries,"

        print(f"{self.cgreen}{msg[:-1]}{self.creset}")

        del Timer.call_stacks[self]



# coding=utf-8
from .query_logging import DjangoSQLWrapper, django_sql_debug
from .query_counting import count_queries

__all__ = [
    "DjangoSQLWrapper",
    "django_sql_debug",
    "count_queries",
]



from collections import defaultdict
from typing import Iterable


def group_by_key(items: Iterable, attribute_name: str) -> dict:
    """
    For a given iterable, returns a dict where each key points to a list
    of items, identified by the passed `attribute_name`.

    Equivalent to:

    ```
    groups = defaultdict(list)
    for item in items:
        groups[item.attribute_name].append(item)
    ```
    """

    grouped = defaultdict(list)
    for item in items:
        attribute = getattr(item, attribute_name)
        if callable(attribute):
            attribute = attribute()
        grouped[attribute].append(item)

    return dict(grouped)



from django.db import connections

total_query_count = 0


class count_queries:
    call_stacks = {}

    def __init__(self, name="", db_name="default"):
        self.name = name
        self.db_name = db_name

    def _get_padding(self):
        return " " * (count_queries.call_stacks[self]) * 3

    def __enter__(self):
        count_queries.call_stacks[self] = len(count_queries.call_stacks.keys())

        con = connections[self.db_name]
        self.queries_start = len(con.queries)

        return con

    def __exit__(self, exc_type, exc_val, exc_tb):
        con = connections[self.db_name]
        amount_queries = len(con.queries) - self.queries_start
        global total_query_count
        total_query_count += amount_queries

        blue = "\33[34m"
        reset = "\33[0m"
        name = f"'{self.name}' " if self.name else ""
        print(f"{self._get_padding()}{blue}{name}took {amount_queries} / {total_query_count} queries{reset}")

        del count_queries.call_stacks[self]



from enum import EnumMeta


class EnumContainsMeta(EnumMeta):
    """
    Constructing Enums with this metaclass will allow for `in` checks.
    Note, that this only checks for the existence in the Enum members,
    not the respective values. Similar to dictionary lookups: `"key" in dict_instance`

    ```
        class AtomEnum(Enum, metaclass=EnumContainsMeta):
            hydrogen = 1
            helium = 2
            lithium = 3
            neon = 10

        "hydrogen" in AtomEnum  # True
        "sulfur" in AtomEnum    # False
        "water" not in AtomEnum  # True
    ```
    """

    def __contains__(cls, item):
        return item in cls.__members__



import pickle
from functools import wraps, partial
from pathlib import Path
from typing import Callable

Empty = object()


def _unpickle(destination: Path):
    try:
        with open(destination, "rb") as f:
            return pickle.load(f)
    except FileNotFoundError:
        return Empty


def pklcache(func: Callable | None = None, **options):
    """
    A decorator to quickly cache a functions return value. It is cached as
    pickled file, and unpickled again on the next invocation.
    """

    if func is None:
        return partial(pklcache, **options)

    force_refresh = options.get("force_refresh", False)
    destination = options.get("destination", None)
    if destination is None:
        destination = func.__name__

    @wraps(func)
    def wrapped(*args, **kwargs):
        dest_path = Path(f"{destination}.pkl")
        dest_path.parent.mkdir(parents=True, exist_ok=True)

        if not force_refresh and (result := _unpickle(dest_path)) is not Empty:
            return result

        result = func(*args, **kwargs)
        with open(dest_path, "wb") as f:
            pickle.dump(result, f)

        return result

    return wrapped



from typing import Callable, Any

from baumbelt.timing import MeasureTime


class HuggingLog:
    timer: MeasureTime
    name: str
    prefix: str
    logging_fn: Callable[[str], Any]

    def __init__(self, name: str, logging_fn: Callable[[str], Any] = print, prefix: str | None = None):
        self.name = name
        self.prefix = f"{prefix}: " if prefix else ""
        self.logging_fn = logging_fn

    def __enter__(self):
        self.logging_fn(f"{self.prefix}Start  '{self.name}'...")
        self.timer = MeasureTime().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.timer:
            return

        self.timer.__exit__(exc_type, exc_val, exc_tb)
        duration_str = f"{self.timer.duration} ({self.timer.duration.total_seconds():f}s total)"
        self.logging_fn(f"{self.prefix}Finish '{self.name}' in {duration_str}")



import pickle
from pathlib import Path
from unittest import TestCase

from src.baumbelt.cache import pklcache


class PklCacheTestCase(TestCase):
    def tearDown(self):
        Path("work.pkl").unlink(missing_ok=True)

    def test_pklcache(self):
        @pklcache
        def work(foo: int, bar: int):
            return foo + bar

        dest = Path("work.pkl")
        self.assertFalse(dest.exists())

        r = work(2, 3)
        self.assertEqual(r, 5)
        self.assertTrue(dest.exists())
        with open(dest, "rb") as f:
            unpickled = pickle.load(f)

        self.assertEqual(unpickled, r)
        dest.unlink()

    def test_pklcache_custom_destination(self):
        @pklcache(destination="tmp/work")
        def work(foo: int, bar: int):
            return foo + bar

        dest = Path("tmp/work.pkl")
        self.assertFalse(dest.exists())

        _ = work(2, 3)
        self.assertTrue(dest.exists())
        self.assertFalse(Path("work.pkl").exists())
        dest.unlink()

    def test_pklcache_force_refresh(self):
        @pklcache
        def work(foo: int, bar: int):
            return foo + bar

        dest = Path("work.pkl")
        _ = work(2, 3)
        r = work(200, 300)
        self.assertEqual(r, 5, "Known limitation: differing arguments dont alter the cached file")

        @pklcache(force_refresh=True)
        def work(foo: int, bar: int):
            return foo + bar

        r = work(200, 300)
        self.assertEqual(
            r,
            500,
            "With force_refresh, cached value should be equal to latest invocation",
        )
        r = work(2, 3)
        self.assertEqual(
            r,
            5,
            "Since pklcache is configured to force_refresh, altered arguments alter the cache",
        )
        dest.unlink()



import contextlib
from collections import defaultdict
from datetime import datetime
from typing import Iterable, TypeVar, TypedDict

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None


FILE_MARKER = "<files>"
CleanName = TypeVar("CleanName", bound=str)


class BucketItem(TypedDict):
    ETag: str
    Key: CleanName
    LastModified: datetime
    Size: int


BucketDict = dict[CleanName, BucketItem]


def prettify_tree(d, indent=0):
    for key, value in d.items():
        if key == FILE_MARKER:
            if value:
                for f in value:
                    print("  " * indent + f)
        else:
            print("  " * indent + str(key))
            if isinstance(value, dict):
                prettify_tree(value, indent + 1)
            else:
                print("  " * (indent + 1) + str(value))


def _attach(branch, trunk):
    parts = branch.split("/", 1)
    if len(parts) == 1:  # branch is a file
        trunk[FILE_MARKER].append(parts[0])
    else:
        node, others = parts
        if node not in trunk:
            trunk[node] = defaultdict(dict)
            trunk[node][FILE_MARKER] = []
        _attach(others, trunk[node])


def build_tree_from_files_list(files: Iterable[str]) -> dict:
    """
    Tree building is heavily inspired by https://stackoverflow.com/a/8496834/2547281

    It builds a dictionary, that mirrors a file tree, based on a list of file paths. Eg:

        ["root/sub1/baum.txt", "root/sub1/sub11/foo.txt", "root/bar.json"]

    yields:

        root:
            <files>: [bar.json]
            sub1:
                <files>: [baum.txt]
                sub11:
                    <files>: [foo.txt]

    """
    directory_tree = defaultdict(list)
    directory_tree[FILE_MARKER] = []

    for filepath in files:
        _attach(filepath, directory_tree)

    return directory_tree


def get_content_at_path(path: str, tree: dict) -> tuple[list[str], list[str]]:
    d = tree
    if path.endswith("/"):
        path = path[:-1]

    for part in path.split("/"):
        if part not in d:
            return [], []
        d = d[part]

    files = d[FILE_MARKER]
    directories = [_d for _d in d if _d != FILE_MARKER]
    return directories, files


def is_dir_in_tree(tree: dict, name: str) -> bool:
    d = tree
    res = None
    for part in name.split("/"):
        if part not in d:
            res = False
            break
        d = d[part]
    if res is None:
        res = True

    return res


def wait_for_tasks(futures: list, desc):
    exceptions = []
    cm = tqdm(total=len(futures), desc=desc) if tqdm else contextlib.nullcontext()
    with cm as pbar:
        for fut in futures:
            try:
                _ = fut.result(timeout=60)
            except Exception as e:
                exceptions.append(e)
            if pbar:
                pbar.update(1)
    return exceptions



import logging
from datetime import datetime, timedelta
from time import sleep

import requests
from requests.adapters import HTTPAdapter, DEFAULT_POOLSIZE, DEFAULT_POOLBLOCK

logger = logging.getLogger(__name__)


class SmartRetryHTTPAdapter(HTTPAdapter):
    """
    Adapter that combines multiple request timeout strategies:
    - all times specified in seconds as float
    - overall_timeout specifies the total time after which an end of the request is guaranteed
    - single_connect_timeout and single_read_timeout specify a timeout for individual requests
    - if individual requests time out or fail with server errors (5XX), they are retried as long as there still time
      left before the overall timeout
    - the number of retries does not matter - if the request fails fast, more retries may fit into the overall timeout
      window
    - a tuple of backoff_times is used between failing requests - if there are more retries than elements, it will
      re-use the last element
    - give_up_threshold specifies the time at which, if the remaining time until the overall timeout falls below it,
      no more retries are attempted (because such a low timeout for a single request would make no sense)
    """

    def __init__(
        self,
        overall_timeout: float = 50.0,
        single_connect_timeout: float = 5.0,
        single_read_timeout: float = 45.0,
        backoff_times: tuple[float] = (0.1, 0.25, 0.5, 1.0, 2.0),
        give_up_threshold: float = 1.0,
        pool_connections=DEFAULT_POOLSIZE,
        pool_maxsize=DEFAULT_POOLSIZE,
        pool_block=DEFAULT_POOLBLOCK,
    ):
        super().__init__(
            pool_connections=pool_connections,
            pool_maxsize=pool_maxsize,
            pool_block=pool_block,
        )

        self.overall_timeout: float = overall_timeout
        self.single_connect_timeout: float = single_connect_timeout
        self.single_read_timeout: float = single_read_timeout
        self.backoff_times: tuple[float] = backoff_times
        self.give_up_threshold: float = give_up_threshold

        self.single_timeout: float = self.single_connect_timeout + self.single_read_timeout
        self.connect_timeout_ratio: float = self.single_connect_timeout / self.single_timeout
        self.min_connect_timeout: float = self.give_up_threshold / 2

    def send(self, request, stream=False, timeout=None, verify=True, cert=None, proxies=None) -> requests.Response:
        start = datetime.now()
        end = start + timedelta(seconds=self.overall_timeout)
        attempts = 0
        last_error = None

        while True:
            time_til_end = end - datetime.now()
            max_timeout = max(time_til_end.total_seconds(), 0.0)
            if max_timeout < self.give_up_threshold:
                logger.debug(
                    f"no time left for another retry, aborting ({max_timeout=:.1f}, {self.give_up_threshold=:.1f})"
                )
                break

            planned_connect_timeout = max_timeout * self.connect_timeout_ratio
            max_connect_timeout = max(planned_connect_timeout, self.min_connect_timeout)
            max_read_timeout = max_timeout - max_connect_timeout

            connect_timeout = min(self.single_connect_timeout, max_connect_timeout)
            read_timeout = min(self.single_read_timeout, max_read_timeout)

            timeout = (connect_timeout, read_timeout)

            try:
                attempts += 1
                response = super().send(request, stream, timeout, verify, cert, proxies)

            except requests.ConnectTimeout as err:
                logger.debug(f"got a connect timeout, possibly retrying ({connect_timeout=:.1f}, {err=})")
                last_error = err

            except requests.ReadTimeout as err:
                logger.debug(f"got a read timeout, possibly retrying ({read_timeout=:.1f}, {err=})")
                last_error = err

            else:
                if (status := response.status_code) >= 500:
                    logger.debug(f"got a server error, possibly retrying ({status=})")
                else:
                    logger.debug(f"request finished ({status=}, {attempts=})")
                    response.raise_for_status()
                    return response

            backoff_index = min(attempts - 1, len(self.backoff_times) - 1)
            backoff = self.backoff_times[backoff_index]

            time_til_end = end - datetime.now()
            max_sleep = max(time_til_end.total_seconds(), 0.0)
            if max_sleep < backoff:
                logger.debug(f"no time left to wait for backoff, aborting ({max_sleep=:.1f}, {backoff=:.1f})")
                break

            sleep(backoff)

        logger.debug(f"request failed, raising error ({attempts=})")
        if last_error:
            raise last_error
        else:
            raise requests.Timeout(request=request)



import re
import time
from datetime import datetime
from unittest import TestCase

from baumbelt.timing import MeasureTime


class MeasureTimeTestCase(TestCase):
    def test_duration(self):
        t0 = datetime.now()

        with MeasureTime() as mt:
            time.sleep(0.3)
            tduration = datetime.now() - t0

        self.assertAlmostEqual(mt.duration.total_seconds(), tduration.total_seconds(), places=2)

    def test_duration_string(self):
        with MeasureTime() as mt:
            time.sleep(0.3)

        formatted_duration = str(mt)
        match = re.search(r"^0:00:00\.3\d+ \(0\.3\d+s\)$", formatted_duration)
        self.assertIsNotNone(match)



import contextlib
import io
from unittest import TestCase

from baumbelt.logs import HuggingLog
from tests.utils import strip_duration_and_seconds


class HuggingLogTestCase(TestCase):
    def test_hugging_log_print_default(self):
        buffer = io.StringIO()
        with contextlib.redirect_stdout(buffer), HuggingLog("cross-compile doom"):
            ...

        msg = strip_duration_and_seconds(buffer.getvalue())
        self.assertEqual(msg, "Start  'cross-compile doom'...\nFinish 'cross-compile doom' in <duration>\n")

    def test_hugging_log_logging_fn(self):
        buffer = ""

        def logging_fn(string):
            nonlocal buffer
            buffer += string

        with HuggingLog("cross-compile doom", logging_fn=logging_fn):
            ...

        msg = strip_duration_and_seconds(buffer)
        self.assertEqual(msg, "Start  'cross-compile doom'...Finish 'cross-compile doom' in <duration>")

    def test_hugging_log_prefix(self):
        buffer = io.StringIO()
        with contextlib.redirect_stdout(buffer), HuggingLog("cross-compile doom", prefix="[ARM]"):
            ...

        msg = strip_duration_and_seconds(buffer.getvalue())
        self.assertEqual(
            msg,
            "[ARM]: Start  'cross-compile doom'...\n[ARM]: Finish 'cross-compile doom' in <duration>\n",
        )



from enum import Enum
from unittest import TestCase

from baumbelt.enums import EnumContainsMeta


class AtomEnum(Enum, metaclass=EnumContainsMeta):
    hydrogen = 1
    helium = 2
    lithium = 3
    neon = 10


class EnumContainsTestCase(TestCase):
    def test_enum_contains(self):
        self.assertIn("hydrogen", AtomEnum)
        self.assertIn("helium", AtomEnum)
        self.assertIn("lithium", AtomEnum)
        self.assertIn("neon", AtomEnum)
        self.assertNotIn("water", AtomEnum)



import logging
import sys
from datetime import timedelta
from time import sleep

from django.conf import settings
from django.core.management import BaseCommand, call_command
from django.utils import timezone

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("--timeout", type=int, default=60)

    def handle(self, *args, **options):
        start = timezone.now()
        timeout_secs = options["timeout"]
        timeout = start + timedelta(seconds=timeout_secs)
        while not self.db_migrated_to_current_state():
            if timezone.now() > timeout:
                logger.warning(f"exceeded timeout ({timeout_secs}s) - db is still not migrated")
                sys.exit(1)
            sleep(2)
        logger.debug(f"ready after {timezone.now() - start} - db is migrated")

    def db_migrated_to_current_state(self) -> bool:
        status_map: dict[str, bool | str] = {}
        for db in settings.DATABASES.keys():
            try:
                call_command("migrate", database=db, interactive=False, check_unapplied=True)
            except (SystemExit, Exception) as exc:
                if isinstance(exc, SystemExit) and exc.code == 1:
                    status_map[db] = False
                else:
                    status_map[db] = f"{type(exc).__name__} - {exc}"
            else:
                status_map[db] = True

        if any(status is not True for status in status_map.values()):
            logger.debug(f"db migration status: {status_map}")
            return False

        return True



from concurrent.futures import ThreadPoolExecutor
from io import BytesIO

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.timezone import make_naive
from storages.backends.s3boto3 import S3Boto3Storage
from storages.utils import clean_name, setting


from .utils import (
    BucketDict,
    CleanName,
    BucketItem,
    build_tree_from_files_list,
    is_dir_in_tree,
    get_content_at_path,
    wait_for_tasks,
)


class DirectoryCheckS3Storage(S3Boto3Storage):
    """
    Subclass to enable check for 'directory' existence.

    Imagine your bucket containing a key "<location>/myfolder/myfile.txt".

    When using the base class, the following will be returned:
        exists("myfolder/myfile.txt") -> True
        exists("myfolder/") -> False
        exists("myfolder") -> False
        exists("myfol") -> False

    When using this class, it looks like this:
        exists("myfolder/myfile.txt") -> True
        exists("myfolder/") -> True
        exists("myfolder") -> True
        exists("myfol") -> False
    """

    def exists(self, name):
        if super().exists(name):
            return True

        name = self._normalize_name(clean_name(name))
        dir_name = name if name.endswith("/") else f"{name}/"
        object_list = self.connection.meta.client.list_objects(
            Bucket=self.bucket_name,
            Prefix=dir_name,
            MaxKeys=1,
        )
        if object_list.get("Contents"):
            # there are files with this directory-like prefix -> consider it a directory
            return True

        return False


class PublicMediaStorage(DirectoryCheckS3Storage):
    @property
    def bucket_name(self):
        if not hasattr(settings, "AWS_PUBLIC_BUCKET"):
            raise ImproperlyConfigured("settings must contain AWS_PUBLIC_BUCKET when using PublicMediaStorage")
        return settings.AWS_PUBLIC_BUCKET

    @property
    def location(self):
        if not hasattr(settings, "PUBLIC_MEDIA_LOCATION"):
            raise ImproperlyConfigured("settings must contain PUBLIC_MEDIA_LOCATION when using PublicMediaStorage")
        return settings.PUBLIC_MEDIA_LOCATION

    @property
    def custom_domain(self):
        if not hasattr(settings, "MEDIA_DOMAIN"):
            raise ImproperlyConfigured("settings must contain MEDIA_DOMAIN when using PublicMediaStorage")
        return settings.MEDIA_DOMAIN


class PrivateMediaStorage(DirectoryCheckS3Storage):
    @property
    def bucket_name(self):
        if not hasattr(settings, "AWS_PRIVATE_BUCKET"):
            raise ImproperlyConfigured("settings must contain AWS_PRIVATE_BUCKET when using PrivateMediaStorage")
        return settings.AWS_PRIVATE_BUCKET

    @property
    def location(self):
        if not hasattr(settings, "PRIVATE_MEDIA_LOCATION"):
            raise ImproperlyConfigured("settings must contain PRIVATE_MEDIA_LOCATION when using PrivateMediaStorage")
        return settings.PRIVATE_MEDIA_LOCATION

    custom_domain = None  # private


class StaticStorage(DirectoryCheckS3Storage):
    querystring_auth = False  # see storages.backends.s3boto3.S3StaticStorage

    @property
    def bucket_name(self):
        if not hasattr(settings, "AWS_PUBLIC_BUCKET"):
            raise ImproperlyConfigured("settings must contain AWS_PUBLIC_BUCKET when using StaticStorage")
        return settings.AWS_PUBLIC_BUCKET

    @property
    def location(self):
        if not hasattr(settings, "STATIC_LOCATION"):
            raise ImproperlyConfigured("settings must contain STATIC_LOCATION when using StaticStorage")
        return settings.STATIC_LOCATION

    @property
    def custom_domain(self):
        if not hasattr(settings, "STATIC_DOMAIN"):
            raise ImproperlyConfigured("settings must contain STATIC_DOMAIN when using StaticStorage")
        return settings.STATIC_DOMAIN


class BulkStaticStorage(StaticStorage):
    """
    When working on many static files, a significant IO overhead is paid.
    To enable a "bulk" strategy with S3, we do the following:

        - defer saving files
        - prefetch bucket contents for fast `exists` and `modified_time` lookups
        - build a local tree-structure of the bucket's contents for `listdir` lookups

    Only on "post_process", which `collectstatic` calls after iterating over
    local files, we spawn a bunch of threads to upload/delete the files.

    To properly support the `--clear` flag, we needed to hack around with the `listdir`
    function.
    """

    # This dict maps absolute file paths (eg `static/admin/css/dropdown.css`) to a
    # BucketItem dictionary, containing meta information.
    _fetched_bucket: BucketDict | None = None
    # The file_tree mirrors a filesystem-structure, with nested dicts mirroring
    # folders containing files and other folders.
    _bucket_file_tree: dict | None = None

    _files_to_save: list[tuple[str, BytesIO]] = []
    _files_to_delete: set[str] = set()

    # >7 threads tend to cause hangs on API calls.
    S3_API_THREADS = 7

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._real_delete_fn = super().delete

    def _fetch_bucket(self):
        client = self.connection.meta.client
        iterator = client.get_paginator("list_objects_v2").paginate(Bucket=self.bucket_name, Prefix=self.location)
        items: BucketDict = {}
        for page in iterator:
            item: BucketItem
            for item in page.get("Contents", []):
                items[item["Key"]] = item

        return items

    @property
    def bucket_contents(self) -> BucketDict:
        if self._fetched_bucket is None:
            self._fetched_bucket = self._fetch_bucket()
        return self._fetched_bucket

    @property
    def bucket_tree(self):
        if self._bucket_file_tree is None:
            self._bucket_file_tree = build_tree_from_files_list(files=self.bucket_contents.keys())
        return self._bucket_file_tree

    def _prepare_name(self, name: str) -> CleanName:
        return self._normalize_name(clean_name(name))

    def exists(self, name: str):
        # Special case for `--clear` and root-directory.
        if name == "":
            return True

        cleaned_name = self._prepare_name(name)
        exists_in_bucket = cleaned_name in self.bucket_contents
        if exists_in_bucket:
            return True

        return is_dir_in_tree(tree=self.bucket_tree[self.location], name=name)

    def listdir(self, name):
        cleaned_name = self._prepare_name(name)
        dirs, files = get_content_at_path(path=cleaned_name, tree=self.bucket_tree)
        return dirs, files

    def get_modified_time(self, name: str):
        cleaned_name = self._prepare_name(name)
        item: BucketItem = self.bucket_contents[cleaned_name]

        # Mirrors super().get_modified_time(name) behaviour.
        if setting("USE_TZ"):
            return item["LastModified"]
        else:
            return make_naive(item["LastModified"])

    def delete(self, name):
        self._files_to_delete.add(name)
        cleaned_name = self._prepare_name(name)
        # Remove our local reference to that file, so that we can act like we never saw it later.
        del self.bucket_contents[cleaned_name]

    def save(self, name, fileobj, max_length=None):
        # have to read the file's content here, otherwise it is closed again once we need it.
        self._files_to_save.append((name, BytesIO(fileobj.read())))

        try:
            # If file is also marked to be deleted on S3, we remove that again.
            # Essentially overwriting what S3 holds. Only relevant in `--clear` scenario.
            self._files_to_delete.remove(name)
        except KeyError:
            pass

    def _upload_deferred_save_calls(self):
        with ThreadPoolExecutor(max_workers=self.S3_API_THREADS) as executor:
            futures = [executor.submit(self._save, name=name, content=content) for name, content in self._files_to_save]

        exceptions = wait_for_tasks(futures, desc="s3 upload")

        print(f"Upload-Executor done, {len(self._files_to_save)} tasks submitted.")
        if exceptions:
            print(f"{len(exceptions)} exceptions occurred. Here's the last one. Enjoy!")
            raise AssertionError("Uploading failed") from exceptions[-1]

    def _flush_deferred_delete_calls(self):
        with ThreadPoolExecutor(max_workers=self.S3_API_THREADS) as executor:
            futures = [executor.submit(self._real_delete_fn, name=name) for name in self._files_to_delete]

        exceptions = wait_for_tasks(futures, desc="s3 delete")

        print(f"Delete-Executor done, {len(self._files_to_save)} tasks submitted.")
        if exceptions:
            print(f"{len(exceptions)} exceptions occurred. Here's the last one. Enjoy!")
            raise AssertionError("Deleting failed") from exceptions[-1]

    def post_process(self, file_paths: list[str], dry_run: bool = False):
        if not dry_run:
            self._upload_deferred_save_calls()
            self._flush_deferred_delete_calls()

        return []


class CdnBulkStaticStorage(BulkStaticStorage):
    """
    When configuring a CDN (eg AWS Cloudfront), it's likely that it serves
    from a private bucket. The default BulkStaticStorage requires a public bucket.
    """

    @property
    def bucket_name(self):
        if not hasattr(settings, "AWS_PRIVATE_BUCKET"):
            raise ImproperlyConfigured("settings must contain AWS_PRIVATE_BUCKET when using CdnBulkStaticStorage")
        return settings.AWS_PRIVATE_BUCKET
